"""The version number is based on semantic versioning.

References
- https://semver.org/
- https://www.python.org/dev/peps/pep-0440/

This file is autogenerated, do not modify by hand.
"""
__version__ = "1.0.0-beta.61"
COMMIT = "0c4c6894bb03ac8bc31aa2767dbbcccf9b23deb5"
MAJOR = 1
MINOR = 0
PATCH = 0



"""List all devices cli command."""
import click
import json
from operator import attrgetter
from typing import Iterable
from tabulate import tabulate

from mbed_devices import get_connected_devices, Device
from mbed_targets import MbedTarget


@click.command()
@click.option(
    "--format", type=click.Choice(["table", "json"]), default="table", show_default=True, help="Set output format."
)
def list_connected_devices(format: str) -> None:
    """Prints connected devices."""
    devices = _sort_devices_by_name(get_connected_devices())
    output_builders = {
        "table": _build_tabular_output,
        "json": _build_json_output,
    }
    if devices:
        output = output_builders[format](devices)
        click.echo(output)
    else:
        click.echo("No connected Mbed devices found.")


def _sort_devices_by_name(devices: Iterable[Device]) -> Iterable[Device]:
    return sorted(devices, key=attrgetter("mbed_target.board_name"))


def _build_tabular_output(devices: Iterable[Device]) -> str:
    headers = ["Board name", "Serial number", "Serial port", "Mount point(s)", "Build target(s)"]
    devices_data = []
    for device in devices:
        devices_data.append(
            [
                device.mbed_target.board_name,
                device.serial_number,
                device.serial_port or "UNKNOWN",
                "\n".join(str(mount_point) for mount_point in device.mount_points),
                "\n".join(_get_build_targets(device.mbed_target)),
            ]
        )
    return tabulate(devices_data, headers=headers)


def _build_json_output(devices: Iterable[Device]) -> str:
    devices_data = []
    for device in devices:
        mbed_target = device.mbed_target
        devices_data.append(
            {
                "serial_number": device.serial_number,
                "serial_port": device.serial_port,
                "mount_points": [str(m) for m in device.mount_points],
                "mbed_target": {
                    "product_code": mbed_target.product_code,
                    "board_type": mbed_target.board_type,
                    "board_name": mbed_target.board_name,
                    "mbed_os_support": mbed_target.mbed_os_support,
                    "mbed_enabled": mbed_target.mbed_enabled,
                    "build_targets": _get_build_targets(mbed_target),
                },
            }
        )
    return json.dumps(devices_data, indent=4)


def _get_build_targets(mbed_target: MbedTarget) -> Iterable[str]:
    return [f"{mbed_target.board_type}_{variant}" for variant in mbed_target.build_variant] + [mbed_target.board_type]



"""Integration with https://github.com/ARMmbed/mbed-tools."""
from mbed_devices._internal.mbed_tools.list_connected_devices import list_connected_devices
from mbed_targets.mbed_tools import config_variables as mbed_targets_config_variables

cli = list_connected_devices
config_variables = mbed_targets_config_variables



"""API for listing devices."""
from typing import Iterable

from mbed_targets.exceptions import MbedTargetsError

from mbed_devices._internal.candidate_device import CandidateDevice
from mbed_devices._internal.detect_candidate_devices import detect_candidate_devices
from mbed_devices._internal.resolve_target import resolve_target
from mbed_devices._internal.exceptions import NoTargetForCandidate

from mbed_devices.device import Device
from mbed_devices.exceptions import DeviceLookupFailed


def get_connected_devices() -> Iterable[Device]:
    """Returns Mbed Devices connected to host computer."""
    devices = []
    for candidate_device in detect_candidate_devices():
        try:
            devices.append(_build_device(candidate_device))
        except NoTargetForCandidate:
            pass
    return devices


def _build_device(candidate: CandidateDevice) -> Device:
    try:
        return Device(
            serial_port=candidate.serial_port,
            serial_number=candidate.serial_number,
            mount_points=candidate.mount_points,
            mbed_target=resolve_target(candidate),
        )
    except MbedTargetsError as err:
        raise DeviceLookupFailed("Failed to find a connected device.") from err



"""Exceptions internal to the package."""

from mbed_tools_lib.exceptions import ToolsError


class SystemException(ToolsError):
    """Exception with regards to the underlying operating system."""


class NoTargetForCandidate(ToolsError):
    """Raised when target cannot be determined for a candidate."""



"""Resolve targets for `CandidateDevice`.

Resolving a target involves looking up an `MbedTarget` from the `mbed-targets` API, using data found in the "htm file"
located on an "Mbed Enabled" device's USB MSD.

For more information on the mbed-targets package visit https://github.com/ARMmbed/mbed-targets
"""
import itertools
import pathlib

from typing import Iterable, List, Optional

from mbed_targets import MbedTarget, get_target_by_product_code, get_target_by_online_id
from mbed_targets.exceptions import UnknownTarget

from mbed_devices._internal.htm_file import OnlineId, read_online_id, read_product_code
from mbed_devices._internal.candidate_device import CandidateDevice
from mbed_devices._internal.exceptions import NoTargetForCandidate


def resolve_target(candidate: CandidateDevice) -> MbedTarget:
    """Resolves target for a given CandidateDevice.

    This function interrogates CandidateDevice, attempting to establish the best method to resolve an MbedTarget,
    the rules are as follows:

    1. Use product code retrieved from one of HTM files in the mass storage if available.
    2. Use online id retrieved from one of the HTM files in the mass storage if available.
    3. Fallback to product code retrieved from serial number.

    The specification of HTM files is that they redirect to devices product page on os.mbed.com.
    Information about Mbed Enabled requirements: https://www.mbed.com/en/about-mbed/mbed-enabled/requirements/
    """
    all_files_contents = _get_all_htm_files_contents(candidate.mount_points)

    product_code = _extract_product_code(all_files_contents)
    if product_code:
        try:
            return get_target_by_product_code(product_code)
        except UnknownTarget:
            raise NoTargetForCandidate

    online_id = _extract_online_id(all_files_contents)
    if online_id:
        try:
            return get_target_by_online_id(slug=online_id.device_slug, target_type=online_id.device_type)
        except UnknownTarget:
            raise NoTargetForCandidate

    # Product code might be the first 4 characters of the serial number
    try:
        return get_target_by_product_code(candidate.serial_number[:4])
    except UnknownTarget:
        raise NoTargetForCandidate


def _extract_product_code(all_files_contents: Iterable[str]) -> Optional[str]:
    """Return first product code found in files contents, None if not found."""
    for contents in all_files_contents:
        product_code = read_product_code(contents)
        if product_code:
            return product_code
    return None


def _extract_online_id(all_files_contents: Iterable[str]) -> Optional[OnlineId]:
    """Return first online id found in files contents, None if not found."""
    for contents in all_files_contents:
        online_id = read_online_id(contents)
        if online_id:
            return online_id
    return None


def _get_all_htm_files_contents(directories: Iterable[pathlib.Path]) -> List[str]:
    """Yields all htm files contents found in the list of given directories."""
    files_in_each_directory = (directory.iterdir() for directory in directories)
    all_files = itertools.chain.from_iterable(files_in_each_directory)
    return [file.read_text() for file in all_files if _is_htm_file(file)]


def _is_htm_file(file: pathlib.Path) -> bool:
    extensions = [".htm", ".HTM"]
    return file.suffix in extensions and not file.name.startswith(".")



"""Public exceptions raised by the package."""
from mbed_tools_lib.exceptions import ToolsError


class MbedDevicesError(ToolsError):
    """Base public exception for the mbed-devices package."""


class DeviceLookupFailed(MbedDevicesError):
    """Failed to look up data associated with the device."""



"""Device model definition."""
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Optional
from mbed_targets import MbedTarget


@dataclass(frozen=True, order=True)
class Device:
    """Definition of an Mbed Enabled Device.

    An Mbed Device is always a USB mass storage device, which sometimes also presents a USB serial port.
    A valid Mbed Device must have an MbedTarget associated with it.

    Attributes:
        mbed_target: The MbedTarget associated with this device.
        serial_number: The serial number presented by the device to the USB subsystem.
        serial_port: The serial port presented by this device, could be None.
        mount_points: The filesystem mount points associated with this device.
    """

    mbed_target: MbedTarget
    serial_number: str
    serial_port: Optional[str]
    mount_points: Tuple[Path, ...]



"""Package definition for PyPI."""
import os

from setuptools import setup

PROJECT_SLUG = "mbed-devices"
SOURCE_DIR = "mbed_devices"
__version__ = None

repository_dir = os.path.dirname(__file__)

# Read package version, this will set the variable `__version__` to the current version.
with open(os.path.join(repository_dir, SOURCE_DIR, "_version.py")) as fh:
    exec(fh.read())

# Use readme needed as long description in PyPI
with open(os.path.join(repository_dir, "README.md")) as fh:
    long_description = fh.read()

setup(
    author="Mbed team",
    author_email="support@mbed.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Embedded Systems",
    ],
    description="This package allows to list Mbed OS devices which are connected to the host computer.",
    keywords="Arm Mbed OS MbedOS devices list",
    include_package_data=True,
    install_requires=[
        "python-dotenv",
        "Click==7.0",
        "mbed-targets",
        "mbed-tools-lib",
        "pywin32; platform_system=='Windows'",
        "psutil; platform_system=='Linux'",
        "pyudev; platform_system=='Linux'",
        "typing-extensions",
        "tabulate",
        "dataclasses; python_version<'3.7'",
    ],
    license="Apache 2.0",
    long_description_content_type="text/markdown",
    long_description=long_description,
    name=PROJECT_SLUG,
    packages=[SOURCE_DIR],
    python_requires=">=3.6,<4",
    url=f"https://github.com/ARMmbed/{PROJECT_SLUG}",
    version=__version__,
)



"""Exposes the primary interfaces for the library."""

from mbed_devices._version import __version__
from mbed_devices.mbed_devices import get_connected_devices
from mbed_devices.device import Device
from mbed_devices import exceptions
